@@ -8,9 +8,7 @@ import asyncio
 import getpass
 import logging
 import socket
-from dataclasses import dataclass, field
 from pathlib import Path
-from shutil import rmtree
 from textwrap import wrap
 
 import click
@@ -18,21 +16,11 @@ import docker
 from packaging.version import parse
 from tabulate import tabulate
 
-from .core import (
-    APPLICATION_ID,
-    DEFAULT_PORT,
-    LOGGER,
-    AiidaLabInstance,
-    Config,
-    Profile,
-)
-from .util import (
-    confirm_with_value,
-    get_docker_client,
-    get_latest_version,
-    spinner,
-    webbrowser_available,
-)
+from .application_state import ApplicationState
+from .core import LOGGER
+from .instance import AiidaLabInstance
+from .profile import DEFAULT_PORT, Profile
+from .util import confirm_with_value, get_latest_version, spinner, webbrowser_available
 from .version import __version__
 
 MSG_MOUNT_POINT_CONFLICT = """Warning: There is at least one other running
@@ -74,24 +62,6 @@ LOGGING_LEVELS = {
 }  #: a mapping of `verbose` option counts to logging levels
 
 
-def _application_config_path():
-    return Path(click.get_app_dir(APPLICATION_ID)) / "config.toml"
-
-
-def _load_config():
-    try:
-        return Config.load(_application_config_path())
-    except FileNotFoundError:
-        return Config()
-
-
-@dataclass
-class ApplicationState:
-
-    config: Config = field(default_factory=_load_config)
-    docker_client: docker.DockerClient = field(default_factory=get_docker_client)
-
-
 pass_app_state = click.make_pass_decorator(ApplicationState, ensure=True)
 
 
@@ -114,7 +84,8 @@ def with_profile(cmd):
     count=True,
     help="Provide this option to increase the output verbosity of the launcher.",
 )
-def cli(verbose):
+@pass_app_state
+def cli(app_state, verbose):
     # Use the verbosity count to determine the logging level...
     logging.basicConfig(
         level=LOGGING_LEVELS[verbose] if verbose in LOGGING_LEVELS else logging.DEBUG
@@ -127,7 +98,7 @@ def cli(verbose):
             err=True,
         )
 
-    LOGGER.info(f"Configuration file path: {_application_config_path()}")
+    LOGGER.info(f"Configuration file path: {app_state.config_path}")
 
     latest_version = get_latest_version(timeout=0.1)
     if latest_version and latest_version > parse(__version__):
@@ -138,6 +109,9 @@ def cli(verbose):
         if "pipx" in __file__:
             click.secho("Run `pipx upgrade aiidalab-launch` to update.", fg="yellow")
 
+    # Apply migrations
+    app_state.apply_migrations()
+
 
 @cli.command()
 def version():
@@ -176,7 +150,7 @@ def list_profiles(app_state):
 @click.argument("profile")
 @pass_app_state
 def show_profile(app_state, profile):
-    """Show a AiiDAlab profile configuration."""
+    """Show an AiiDAlab profile configuration."""
     click.echo(app_state.config.get_profile(profile).dumps(), nl=False)
 
 
@@ -210,8 +184,6 @@ def add_profile(ctx, app_state, port, home_mount, profile):
     # Determine next available port or use the one provided by the user.
     configured_ports = [prof.port for prof in app_state.config.profiles if prof.port]
     port = port or (max(configured_ports, default=-1) + 1) or DEFAULT_PORT
-    # Determine home mount path unless provided.
-    home_mount = home_mount or str(Path.home() / f"aiidalab-{profile}")
 
     try:
         new_profile = Profile(
@@ -223,7 +195,7 @@ def add_profile(ctx, app_state, port, home_mount, profile):
         raise click.ClickException(error)
 
     app_state.config.profiles.append(new_profile)
-    app_state.config.save(_application_config_path())
+    app_state.save_config()
     click.echo(f"Added profile '{profile}'.")
     if click.confirm("Do you want to edit it now?", default=True):
         ctx.invoke(edit_profile, profile=profile)
@@ -235,7 +207,7 @@ def add_profile(ctx, app_state, port, home_mount, profile):
 @click.option("-f", "--force", is_flag=True, help="Proceed, ignoring any warnings.")
 @pass_app_state
 def remove_profile(app_state, profile, yes, force):
-    """Remove a AiiDAlab profile from the configuration."""
+    """Remove an AiiDAlab profile from the configuration."""
     try:
         profile = app_state.config.get_profile(profile)
     except ValueError:
@@ -259,7 +231,7 @@ def remove_profile(app_state, profile, yes, force):
             f"Are you sure you want to remove profile '{profile.name}'?"
         ):
             app_state.config.profiles.remove(profile)
-            app_state.config.save(_application_config_path())
+            app_state.save_config()
             click.echo(f"Removed profile with name '{profile.name}'.")
 
 
@@ -267,7 +239,7 @@ def remove_profile(app_state, profile, yes, force):
 @click.argument("profile")
 @pass_app_state
 def edit_profile(app_state, profile):
-    """Edit a AiiDAlab profile configuration."""
+    """Edit an AiiDAlab profile configuration."""
     current_profile = app_state.config.get_profile(profile)
     profile_edit = click.edit(current_profile.dumps(), extension=".toml")
     if profile_edit:
@@ -275,7 +247,7 @@ def edit_profile(app_state, profile):
         if new_profile != current_profile:
             app_state.config.profiles.remove(current_profile)
             app_state.config.profiles.append(new_profile)
-            app_state.config.save(_application_config_path())
+            app_state.save_config()
             return
     click.echo("No changes.")
 
@@ -284,14 +256,14 @@ def edit_profile(app_state, profile):
 @click.argument("profile")
 @pass_app_state
 def set_default_profile(app_state, profile):
-    """Set a AiiDAlab profile as default."""
+    """Set an AiiDAlab profile as default."""
     try:
         app_state.config.get_profile(profile)
     except ValueError:
         raise click.ClickException(f"A profile with name '{profile}' does not exist.")
     else:
         app_state.config.default_profile = profile
-        app_state.config.save(_application_config_path())
+        app_state.save_config()
         click.echo(f"Set default profile to '{profile}'.")
 
 
@@ -357,44 +329,41 @@ async def _async_start(
             "Try to use '--pull' to pull the image prior to start."
         )
 
-    # Check if image has changed.
-    assert instance.image is not None
-    image_changed = (
-        instance.container and instance.container.image.id != instance.image.id
-    )
-    mounts_changed = (
-        instance.container
-        and f"{instance.profile.container_name()}_conda"
-        not in (mount.get("Name") for mount in instance.container.attrs["Mounts"])
-    )
-    recreate = image_changed or mounts_changed
+    # Check if the container configuration has changed.
+    if instance.container:
+        configuration_changed = any(instance.configuration_changes())
+    else:
+        configuration_changed = False
 
     try:
 
         InstanceStatus = instance.AiidaLabInstanceStatus  # local alias for brevity
 
         status = await instance.status()
+
+        # Container needs to be started.
         if status in (
             InstanceStatus.DOWN,
             InstanceStatus.CREATED,
             InstanceStatus.EXITED,
         ):
-            if recreate:
+            if configuration_changed:
                 with spinner("Recreating container..."):
-                    instance.remove()
-                    instance.create()
+                    instance.recreate()
             with spinner("Starting container..."):
                 instance.start()
+
+        # Container is already up.
         elif status is InstanceStatus.UP and restart:
             with spinner("Restarting container..."):
-                if any(instance.configuration_changes()):
+                if configuration_changed:
                     instance.stop()
-                    instance.remove()
+                    instance.recreate()
                     instance.start()
                 else:
                     instance.restart()
         elif status is InstanceStatus.UP and not restart:
-            if any(instance.configuration_changes()):
+            if configuration_changed:
                 click.secho(
                     "Container is already running, however the configuration "
                     "has changed. A restart with --restart is recommended.",
@@ -405,9 +374,17 @@ async def _async_start(
                     "Container was already running, use --restart to restart it.",
                     err=True,
                 )
+
+        # Container is already starting.
         elif status is InstanceStatus.STARTING:
             click.echo("Container is already starting up...", err=True)
 
+        # Unknown condition.
+        else:
+            raise RuntimeError(
+                "Container already exists, but failed to determine status."
+            )
+
     except docker.errors.APIError as error:
         LOGGER.debug(f"Error during startup: {error}")
         if instance.profile.port and "port is already allocated" in str(error):
@@ -429,6 +406,7 @@ async def _async_start(
                     echo_logs = asyncio.create_task(instance.echo_logs())
                     await asyncio.wait_for(instance.wait_for_services(), timeout=wait)
                     echo_logs.cancel()
+                    LOGGER.debug("AiiDAlab instance ready.")
             except asyncio.TimeoutError:
                 raise click.ClickException(
                     f"AiiDAlab instance did not start up within the provided wait period ({wait})."
@@ -439,19 +417,22 @@ async def _async_start(
                     "the container output logs by increasing the output "
                     "verbosity with 'aiidalab-launch -vvv start'."
                 )
+            LOGGER.debug("Preparing startup message.")
             msg_startup = (
                 MSG_STARTUP_SSH
                 if (show_ssh_help or not webbrowser_available())
                 else MSG_STARTUP
             )
             url = instance.url()
+            host_ports = instance.host_ports()
+            assert len(host_ports) > 0
             click.secho(
                 msg_startup.format(
                     url=url,
                     home_mount=instance.profile.home_mount,
                     system_user=instance.profile.system_user,
                     user=getpass.getuser(),
-                    port=instance.profile.port,
+                    port=host_ports[0],
                     hostname=socket.getfqdn(),
                 ).lstrip(),
                 fg="green",
@@ -615,7 +596,7 @@ def status(app_state):
 @click.pass_context
 @with_profile
 def exec(ctx, profile, cmd, privileged, forward_exit_code, wait):
-    """Directly execute a command on a AiiDAlab instance.
+    """Directly execute a command on an AiiDAlab instance.
 
     For example, to get a list of all installed aiidalab applications, run:
 
@@ -646,16 +627,15 @@ def exec(ctx, profile, cmd, privileged, forward_exit_code, wait):
 
 
 @cli.command()
-@click.option(
-    "--apps",
-    is_flag=True,
-    help="Only remove installed apps. The default apps will be installed on the next instance start.",
-)
 @click.option("--yes", is_flag=True, help="Do not ask for confirmation.")
 @with_profile
 @pass_app_state
-def reset(app_state, profile, apps, yes):
-    """Reset an AiiDAlab instance."""
+def reset(app_state, profile, yes):
+    """Reset an AiiDAlab instance.
+
+    This function removes all user-installed apps and generated data. Use with
+    caution.
+    """
     # Check (and abort) in case that the instance is running.
     instance = AiidaLabInstance(client=app_state.docker_client, profile=profile)
     status = asyncio.run(instance.status())
@@ -680,26 +660,8 @@ def reset(app_state, profile, apps, yes):
             profile.name, "Please enter the name of the profile to continue", abort=True
         )
 
-    def rmtree_(path: Path) -> None:
-        if path.exists():
-            try:
-                rmtree(path)
-            except Exception as error:
-                raise click.ClickException(
-                    f"Encountered error while trying to remove '{path}': {error}"
-                )
-
-    if apps:
-        click.echo(
-            "Removing apps directory. Default apps will be installed on next start."
-        )
-        rmtree_(Path(profile.home_dir) / "apps")
-    else:
-        click.echo("Removing container and associated volumes.")
-        instance.remove()
-
-        click.echo(f"Removing home directory '{profile.home_mount}'.")
-        rmtree_(Path(profile.home_mount))
+    click.echo("Removing container and associated (data) volumes.")
+    instance.remove(conda=True, data=True)
 
 
 if __name__ == "__main__":
