@@ -1,11 +1,11 @@
-# -*- coding: utf-8 -*-
 import asyncio
-import logging
+import re
 import webbrowser
 from contextlib import contextmanager
+from pathlib import Path, PosixPath, PurePosixPath, WindowsPath
 from textwrap import wrap
 from threading import Event, Thread, Timer
-from typing import Any, AsyncGenerator, Generator, Iterable, Optional
+from typing import Any, AsyncGenerator, Generator, Iterable, Optional, Union
 
 import click
 import click_spinner
@@ -14,6 +14,8 @@ import requests
 from packaging.version import Version, parse
 from requests_cache import CachedSession
 
+from .core import LOGGER
+
 MSG_UNABLE_TO_COMMUNICATE_WITH_CLIENT = (
     "Unable to communicate with docker on this host. This error usually indicates "
     "that Docker is either not installed on this system, that the docker service is "
@@ -133,19 +135,19 @@ def get_latest_version(timeout: float = 0.1) -> Optional[Version]:
         )
         req.raise_for_status()
         releases = sorted(
-            [
+            (
                 version
                 for version, release in req.json()["releases"].items()
                 if not all(r["yanked"] for r in release)
-            ],
+            ),
             key=parse,
         )
         return parse(releases[-1]) if releases else None
     except (requests.exceptions.Timeout, requests.exceptions.ReadTimeout):
-        logging.debug("Timed out while requesting latest version.")
+        LOGGER.debug("Timed out while requesting latest version.")
         return None
     except OSError as error:
-        logging.debug(f"Error while requesting latest version: {error}")
+        LOGGER.debug(f"Error while requesting latest version: {error}")
         return None
 
 
@@ -156,3 +158,47 @@ def confirm_with_value(value: str, text: str, abort: bool = False) -> bool:
         raise click.Abort
     else:
         return False
+
+
+def docker_mount_for(
+    container: docker.models.containers.Container, destination: PurePosixPath
+) -> Union[Path, str]:
+    """Identify the Docker mount bind path or volume for a given destination."""
+    try:
+        mount = [
+            mount
+            for mount in container.attrs["Mounts"]
+            if mount["Destination"] == str(destination)
+        ][0]
+    except IndexError:
+        raise ValueError(f"No mount point for {destination}.")
+    if mount["Type"] == "bind":
+        docker_root = PurePosixPath("/host_mnt")
+        docker_path = PurePosixPath(mount["Source"])
+        try:  # Windows
+            drive = docker_path.relative_to(docker_root).parts[0]
+            return WindowsPath(
+                f"{drive}:",
+                docker_path.root,
+                docker_path.relative_to(docker_root, drive),
+            )
+        except ValueError:  # Linux
+            return PosixPath(docker_path)
+        except NotImplementedError:  # OS-X
+            return PosixPath(docker_root.root, docker_path.relative_to(docker_root))
+    elif mount["Type"] == "volume":
+        return mount["Name"]
+    else:
+        raise RuntimeError("Unexpected mount type.")
+
+
+def get_docker_env(container: docker.models.containers.Container, env_name: str) -> str:
+    re_pattern = f"{re.escape(env_name)}=(?P<value>.*)"
+    try:
+        for item in container.attrs["Config"]["Env"]:
+            match = re.search(re_pattern, item)
+            if match:
+                return match.groupdict()["value"]
+    except KeyError:
+        pass
+    raise KeyError(env_name)
